@@ -20,71 +20,14 @@
 #include "kernel/yosys.h"
 #include "kernel/satgen.h"
 #include "kernel/json.h"
+#include "kernel/gzip.h"
+#include "kernel/log_help.h"
 
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <errno.h>
 
-#ifdef YOSYS_ENABLE_ZLIB
-#include <zlib.h>
-
-PRIVATE_NAMESPACE_BEGIN
-#define GZ_BUFFER_SIZE 8192
-void decompress_gzip(const std::string &filename, std::stringstream &out)
-{
-	char buffer[GZ_BUFFER_SIZE];
-	int bytes_read;
-	gzFile gzf = gzopen(filename.c_str(), "rb");
-	while(!gzeof(gzf)) {
-		bytes_read = gzread(gzf, reinterpret_cast<void *>(buffer), GZ_BUFFER_SIZE);
-		out.write(buffer, bytes_read);
-	}
-	gzclose(gzf);
-}
-
-/*
-An output stream that uses a stringbuf to buffer data internally,
-using zlib to write gzip-compressed data every time the stream is flushed.
-*/
-class gzip_ostream : public std::ostream  {
-public:
-	gzip_ostream() : std::ostream(nullptr)
-	{
-		rdbuf(&outbuf);
-	}
-	bool open(const std::string &filename)
-	{
-		return outbuf.open(filename);
-	}
-private:
-	class gzip_streambuf : public std::stringbuf {
-	public:
-		gzip_streambuf() { };
-		bool open(const std::string &filename)
-		{
-			gzf = gzopen(filename.c_str(), "wb");
-			return gzf != nullptr;
-		}
-		virtual int sync() override
-		{
-			gzwrite(gzf, reinterpret_cast<const void *>(str().c_str()), unsigned(str().size()));
-			str("");
-			return 0;
-		}
-		virtual ~gzip_streambuf()
-		{
-			sync();
-			gzclose(gzf);
-		}
-	private:
-		gzFile gzf = nullptr;
-	} outbuf;
-};
-PRIVATE_NAMESPACE_END
-
-#endif
-
 YOSYS_NAMESPACE_BEGIN
 
 #define MAX_REG_COUNT 1000
@@ -99,7 +42,8 @@ std::map<std::string, Backend*> backend_register;
 
 std::vector<std::string> Frontend::next_args;
 
-Pass::Pass(std::string name, std::string short_help) : pass_name(name), short_help(short_help)
+Pass::Pass(std::string name, std::string short_help, source_location location) : 
+	pass_name(name), short_help(short_help), location(location)
 {
 	next_queued_pass = first_queued_pass;
 	first_queued_pass = this;
@@ -174,9 +118,19 @@ void Pass::post_execute(Pass::pre_post_exec_state_t state)
 
 void Pass::help()
 {
-	log("\n");
-	log("No help message for command `%s'.\n", pass_name.c_str());
-	log("\n");
+	auto prettyHelp = PrettyHelp();
+	if (formatted_help()) {
+		prettyHelp.log_help();
+	} else {
+		log("\n");
+		log("No help message for command `%s'.\n", pass_name.c_str());
+		log("\n");
+	}
+}
+
+bool Pass::formatted_help()
+{
+	return false;
 }
 
 void Pass::clear_flags()
@@ -318,18 +272,18 @@ void Pass::call(RTLIL::Design *design, std::vector<std::string> args)
 	pass_register[args[0]]->execute(args, design);
 	pass_register[args[0]]->post_execute(state);
 	while (design->selection_stack.size() > orig_sel_stack_pos)
-		design->selection_stack.pop_back();
+		design->pop_selection();
 }
 
 void Pass::call_on_selection(RTLIL::Design *design, const RTLIL::Selection &selection, std::string command)
 {
 	std::string backup_selected_active_module = design->selected_active_module;
 	design->selected_active_module.clear();
-	design->selection_stack.push_back(selection);
+	design->push_selection(selection);
 
 	Pass::call(design, command);
 
-	design->selection_stack.pop_back();
+	design->pop_selection();
 	design->selected_active_module = backup_selected_active_module;
 }
 
@@ -337,11 +291,11 @@ void Pass::call_on_selection(RTLIL::Design *design, const RTLIL::Selection &sele
 {
 	std::string backup_selected_active_module = design->selected_active_module;
 	design->selected_active_module.clear();
-	design->selection_stack.push_back(selection);
+	design->push_selection(selection);
 
 	Pass::call(design, args);
 
-	design->selection_stack.pop_back();
+	design->pop_selection();
 	design->selected_active_module = backup_selected_active_module;
 }
 
@@ -349,12 +303,12 @@ void Pass::call_on_module(RTLIL::Design *design, RTLIL::Module *module, std::str
 {
 	std::string backup_selected_active_module = design->selected_active_module;
 	design->selected_active_module = module->name.str();
-	design->selection_stack.push_back(RTLIL::Selection(false));
-	design->selection_stack.back().select(module);
+	design->push_empty_selection();
+	design->select(module);
 
 	Pass::call(design, command);
 
-	design->selection_stack.pop_back();
+	design->pop_selection();
 	design->selected_active_module = backup_selected_active_module;
 }
 
@@ -362,12 +316,12 @@ void Pass::call_on_module(RTLIL::Design *design, RTLIL::Module *module, std::vec
 {
 	std::string backup_selected_active_module = design->selected_active_module;
 	design->selected_active_module = module->name.str();
-	design->selection_stack.push_back(RTLIL::Selection(false));
-	design->selection_stack.back().select(module);
+	design->push_empty_selection();
+	design->select(module);
 
 	Pass::call(design, args);
 
-	design->selection_stack.pop_back();
+	design->pop_selection();
 	design->selected_active_module = backup_selected_active_module;
 }
 
@@ -439,8 +393,8 @@ void ScriptPass::help_script()
 	script();
 }
 
-Frontend::Frontend(std::string name, std::string short_help) :
-		Pass(name.rfind("=", 0) == 0 ? name.substr(1) : "read_" + name, short_help),
+Frontend::Frontend(std::string name, std::string short_help, source_location location) :
+		Pass(name.rfind("=", 0) == 0 ? name.substr(1) : "read_" + name, short_help, location),
 		frontend_name(name.rfind("=", 0) == 0 ? name.substr(1) : name)
 {
 }
@@ -527,47 +481,9 @@ void Frontend::extra_args(std::istream *&f, std::string &filename, std::vector<s
 				next_args.insert(next_args.end(), args.begin(), args.begin()+argidx);
 				next_args.insert(next_args.end(), filenames.begin()+1, filenames.end());
 			}
-			std::ifstream *ff = new std::ifstream;
-			ff->open(filename.c_str(), bin_input ? std::ifstream::binary : std::ifstream::in);
 			yosys_input_files.insert(filename);
-			if (ff->fail()) {
-				delete ff;
-				ff = nullptr;
-			}
-			f = ff;
-			if (f != NULL) {
-				// Check for gzip magic
-				unsigned char magic[3];
-				int n = 0;
-				while (n < 3)
-				{
-					int c = ff->get();
-					if (c != EOF) {
-						magic[n] = (unsigned char) c;
-					}
-					n++;
-				}
-				if (n == 3 && magic[0] == 0x1f && magic[1] == 0x8b) {
-	#ifdef YOSYS_ENABLE_ZLIB
-					log("Found gzip magic in file `%s', decompressing using zlib.\n", filename.c_str());
-					if (magic[2] != 8)
-						log_cmd_error("gzip file `%s' uses unsupported compression type %02x\n",
-							filename.c_str(), unsigned(magic[2]));
-					delete ff;
-					std::stringstream *df = new std::stringstream();
-					decompress_gzip(filename, *df);
-					f = df;
-	#else
-					log_cmd_error("File `%s' is a gzip file, but Yosys is compiled without zlib.\n", filename.c_str());
-	#endif
-				} else {
-					ff->clear();
-					ff->seekg(0, std::ios::beg);
-				}
-			}
+			f = uncompressed(filename, bin_input ? std::ifstream::binary : std::ifstream::in);
 		}
-		if (f == NULL)
-			log_cmd_error("Can't open input file `%s' for reading: %s\n", filename.c_str(), strerror(errno));
 
 		for (size_t i = argidx+1; i < args.size(); i++)
 			if (args[i].compare(0, 1, "-") == 0)
@@ -623,8 +539,8 @@ void Frontend::frontend_call(RTLIL::Design *design, std::istream *f, std::string
 	}
 }
 
-Backend::Backend(std::string name, std::string short_help) :
-		Pass(name.rfind("=", 0) == 0 ? name.substr(1) : "write_" + name, short_help),
+Backend::Backend(std::string name, std::string short_help, source_location location) :
+		Pass(name.rfind("=", 0) == 0 ? name.substr(1) : "write_" + name, short_help, location),
 		backend_name(name.rfind("=", 0) == 0 ? name.substr(1) : name)
 {
 }
@@ -745,7 +661,7 @@ void Backend::backend_call(RTLIL::Design *design, std::ostream *f, std::string f
 	}
 
 	while (design->selection_stack.size() > orig_sel_stack_pos)
-		design->selection_stack.pop_back();
+		design->pop_selection();
 }
 
 struct SimHelper {
@@ -777,6 +693,23 @@ static string get_cell_name(string name) {
 	return is_code_getter(name) ? name.substr(0, name.length()-1) : name;
 }
 
+static void log_warning_flags(Pass *pass) {
+	bool has_warnings = false;
+	const string name = pass->pass_name;
+	if (pass->experimental_flag) {
+		if (!has_warnings) log("\n");
+		has_warnings = true;
+		log("WARNING: THE '%s' COMMAND IS EXPERIMENTAL.\n", name.c_str());
+	}
+	if (pass->internal_flag) {
+		if (!has_warnings) log("\n");
+		has_warnings = true;
+		log("WARNING: THE '%s' COMMAND IS INTENDED FOR INTERNAL DEVELOPER USE ONLY.\n", name.c_str());
+	}
+	if (has_warnings)
+		log("\n");
+}
+
 static struct CellHelpMessages {
 	dict<string, SimHelper> cell_help;
 	CellHelpMessages() {
@@ -802,161 +735,217 @@ struct HelpPass : public Pass {
 		log("    help <celltype>+  ....  print verilog code for given cell type\n");
 		log("\n");
 	}
-	void write_cmd_rst(std::string cmd, std::string title, std::string text)
-	{
-		FILE *f = fopen(stringf("docs/source/cmd/%s.rst", cmd.c_str()).c_str(), "wt");
-		// make header
-		size_t char_len = cmd.length() + 3 + title.length();
-		std::string title_line = "\n";
-		title_line.insert(0, char_len, '=');
-		fprintf(f, "%s", title_line.c_str());
-		fprintf(f, "%s - %s\n", cmd.c_str(), title.c_str());
-		fprintf(f, "%s\n", title_line.c_str());
+	bool dump_cmds_json(PrettyJson &json) {
+		// init json
+		json.begin_object();
+		json.entry("version", "Yosys command reference");
+		json.entry("generator", yosys_version_str);
 
-		// render html
-		fprintf(f, ".. cmd:def:: %s\n", cmd.c_str());
-		fprintf(f, "   :title: %s\n\n", title.c_str());
-		fprintf(f, "   .. only:: html\n\n");
-		std::stringstream ss;
-		std::string textcp = text;
-		ss << text;
-		bool IsUsage = true;
-		int blank_count = 0;
-		size_t def_strip_count = 0;
-		bool WasDefinition = false;
-		for (std::string line; std::getline(ss, line, '\n');) {
-			// find position of first non space character
-			std::size_t first_pos = line.find_first_not_of(" \t");
-			std::size_t last_pos = line.find_last_not_of(" \t");
-			if (first_pos == std::string::npos) {
-				// skip formatting empty lines
-				if (!WasDefinition)
-					fputc('\n', f);
-				blank_count += 1;
-				continue;
+		bool raise_error = false;
+		std::map<string, vector<string>> groups;
+
+		json.name("cmds"); json.begin_object();
+		// iterate over commands
+		for (auto &it : pass_register) {
+			auto name = it.first;
+			auto pass = it.second;
+			auto title = pass->short_help;
+
+			auto cmd_help = PrettyHelp();
+			auto has_pretty_help = pass->formatted_help();
+
+			if (!has_pretty_help) {
+				enum PassUsageState {
+					PUState_none,
+					PUState_signature,
+					PUState_options,
+					PUState_optionbody,
+				};
+
+				source_location null_source;
+				string current_buffer = "";
+				auto root_listing = cmd_help.get_root();
+				auto current_listing = root_listing;
+
+				// dump command help
+				std::ostringstream buf;
+				log_streams.push_back(&buf);
+				pass->help();
+				log_streams.pop_back();
+				std::stringstream ss;
+				ss << buf.str();
+
+				// parse command help
+				size_t def_strip_count = 0;
+				auto current_state = PUState_none;
+				auto catch_verific = false;
+				auto blank_lines = 2;
+				for (string line; std::getline(ss, line, '\n');) {
+					// find position of first non space character
+					std::size_t first_pos = line.find_first_not_of(" \t");
+					std::size_t last_pos = line.find_last_not_of(" \t");
+					if (first_pos == std::string::npos) {
+						switch (current_state)
+						{
+						case PUState_signature:
+							root_listing->usage(current_buffer, null_source);
+							current_listing = root_listing;
+							current_state = PUState_none;
+							current_buffer = "";
+							break;
+						case PUState_none:
+						case PUState_optionbody:
+							blank_lines += 1;
+							break;
+						default:
+							break;
+						}
+						// skip empty lines
+						continue;
+					}
+
+					// strip leading and trailing whitespace
+					std::string stripped_line = line.substr(first_pos, last_pos - first_pos +1);
+					bool IsDefinition = stripped_line[0] == '-';
+					IsDefinition &= stripped_line[1] != ' ' && stripped_line[1] != '>';
+					bool IsDedent = def_strip_count && first_pos < def_strip_count;
+					bool IsIndent = def_strip_count < first_pos;
+
+					// line looks like a signature
+					bool IsSignature = stripped_line.find(name) == 0 && (stripped_line.length() == name.length() || stripped_line.at(name.size()) == ' ');
+
+					if (IsSignature && first_pos <= 4 && (blank_lines >= 2 || current_state == PUState_signature)) {
+						if (current_state == PUState_options || current_state == PUState_optionbody) {
+							current_listing->codeblock(current_buffer, "none", null_source);
+							current_buffer = "";
+						} else if (current_state == PUState_signature) {
+							root_listing->usage(current_buffer, null_source);
+							current_buffer = "";
+						} else if (current_state == PUState_none && !current_buffer.empty()) {
+							current_listing->codeblock(current_buffer, "none", null_source);
+							current_buffer = "";
+						}
+						current_listing = root_listing;
+						current_state = PUState_signature;
+						def_strip_count = first_pos;
+						catch_verific = false;
+					} else if (IsDedent) {
+						def_strip_count = first_pos;
+						if (current_state == PUState_optionbody) {
+							if (!current_buffer.empty()) {
+								current_listing->codeblock(current_buffer, "none", null_source);
+								current_buffer = "";
+							}
+							if (IsIndent) {
+								current_state = PUState_options;
+								current_listing = root_listing->back();
+							} else {
+								current_state = PUState_none;
+								current_listing = root_listing;
+							}
+						} else {
+							current_state = PUState_none;
+						}
+					}
+
+					if (IsDefinition && !catch_verific && current_state != PUState_signature) {
+						if (!current_buffer.empty()) {
+							current_listing->codeblock(current_buffer, "none", null_source);
+							current_buffer = "";
+						}
+						current_state = PUState_options;
+						current_listing = root_listing->open_option(stripped_line, null_source);
+						def_strip_count = first_pos;
+					} else {
+						if (current_state == PUState_options) {
+							current_state = PUState_optionbody;
+						}
+						if (current_buffer.empty())
+							current_buffer = stripped_line;
+						else if (current_state == PUState_signature && IsIndent)
+							current_buffer += stripped_line;
+						else if (current_state == PUState_none) {
+							current_buffer += (blank_lines > 0 ? "\n\n" : "\n") + line;
+						} else
+							current_buffer += (blank_lines > 0 ? "\n\n" : "\n") + stripped_line;
+						if (stripped_line.compare("Command file parser supports following commands in file:") == 0)
+							catch_verific = true;
+					}
+					blank_lines = 0;
+				}
+
+				if (!current_buffer.empty()) {
+					if (current_buffer.size() > 64 && current_buffer.substr(0, 64).compare("The following commands are executed by this synthesis command:\n\n") == 0) {
+						current_listing->paragraph(current_buffer.substr(0, 62), null_source);
+						current_listing->codeblock(current_buffer.substr(64), "yoscrypt", null_source);
+					} else
+						current_listing->codeblock(current_buffer, "none", null_source);
+					current_buffer = "";
+				}
 			}
 
-			// strip leading and trailing whitespace
-			std::string stripped_line = line.substr(first_pos, last_pos - first_pos +1);
-			bool IsDefinition = stripped_line[0] == '-';
-			IsDefinition &= stripped_line[1] != ' ' && stripped_line[1] != '>';
-			bool IsDedent = def_strip_count && first_pos <= def_strip_count;
-			bool IsIndent = first_pos == 2 || first_pos == 4;
-			if (cmd.compare(0, 7, "verific") == 0)
-				// verific.cc has strange and different formatting from the rest
-				IsIndent = false;
-
-			// another usage block
-			bool NewUsage = stripped_line.find(cmd) == 0;
-
-			if (IsUsage) {
-				if (stripped_line.compare(0, 4, "See ") == 0) {
-					// description refers to another function
-					fprintf(f, "\n      %s\n", stripped_line.c_str());
-				} else {
-					// usage should be the first line of help output
-					fprintf(f, "\n      .. code:: yoscrypt\n\n         %s\n\n      ", stripped_line.c_str());
-					WasDefinition = true;
+			// attempt auto group
+			if (!cmd_help.has_group()) {
+				string source_file = pass->location.file_name();
+				bool has_source = source_file.compare("unknown") != 0;
+				if (pass->internal_flag)
+					cmd_help.group = "internal";
+				else if (source_file.find("backends/") == 0 || (!has_source && name.find("read_") == 0))
+					cmd_help.group = "backends";
+				else if (source_file.find("frontends/") == 0 || (!has_source && name.find("write_") == 0))
+					cmd_help.group = "frontends";
+				else if (has_source) {
+					auto last_slash = source_file.find_last_of('/');
+					if (last_slash != string::npos) {
+						auto parent_path = source_file.substr(0, last_slash);
+						cmd_help.group = parent_path;
+					}
 				}
-				IsUsage = false;
-			} else if (IsIndent && NewUsage && (blank_count >= 2 || WasDefinition)) {
-				// another usage block
-				fprintf(f, "\n      .. code:: yoscrypt\n\n         %s\n\n      ", stripped_line.c_str());
-				WasDefinition = true;
-				def_strip_count = 0;
-			} else if (IsIndent && IsDefinition && (blank_count || WasDefinition)) {
-				// format definition term
-				fprintf(f, "\n\n      .. code:: yoscrypt\n\n         %s\n\n      ", stripped_line.c_str());
-				WasDefinition = true;
-				def_strip_count = first_pos;
-			} else {
-				if (IsDedent) {
-					fprintf(f, "\n\n      ::\n");
-					def_strip_count = first_pos;
-				} else if (WasDefinition) {
-					fprintf(f, "::\n");
-					WasDefinition = false;
-				}
-				fprintf(f, "\n         %s", line.substr(def_strip_count, std::string::npos).c_str());
+				// implicit !has_source
+				else if (name.find("equiv") == 0)
+					cmd_help.group = "passes/equiv";
+				else if (name.find("fsm") == 0)
+					cmd_help.group = "passes/fsm";
+				else if (name.find("memory") == 0)
+					cmd_help.group = "passes/memory";
+				else if (name.find("opt") == 0)
+					cmd_help.group = "passes/opt";
+				else if (name.find("proc") == 0)
+					cmd_help.group = "passes/proc";
 			}
 
-			blank_count = 0;
+			if (groups.count(cmd_help.group) == 0) {
+				groups[cmd_help.group] = vector<string>();
+			}
+			groups[cmd_help.group].push_back(name);
+
+			// write to json
+			json.name(name.c_str()); json.begin_object();
+			json.entry("title", title);
+			json.name("content"); json.begin_array();
+			for (auto &content : cmd_help)
+				json.value(content.to_json());
+			json.end_array();
+			json.entry("group", cmd_help.group);
+			json.entry("source_file", pass->location.file_name());
+			json.entry("source_line", pass->location.line());
+			json.entry("source_func", pass->location.function_name());
+			json.entry("experimental_flag", pass->experimental_flag);
+			json.entry("internal_flag", pass->internal_flag);
+			json.end_object();
 		}
-		fputc('\n', f);
+		json.end_object();
 
-		// render latex
-		fprintf(f, ".. only:: latex\n\n");
-		fprintf(f, "   ::\n\n");
-		std::stringstream ss2;
-		ss2 << textcp;
-		for (std::string line; std::getline(ss2, line, '\n');) {
-			fprintf(f, "      %s\n", line.c_str());
-		}
-		fclose(f);
-	}
-	void write_cell_rst(Yosys::SimHelper cell, Yosys::CellType ct)
-	{
-		// open
-		FILE *f = fopen(stringf("docs/source/cell/%s.rst", cell.filesafe_name().c_str()).c_str(), "wt");
+		json.entry("groups", groups);
 
-		// make header
-		string title_line;
-		if (cell.title.length())
-			title_line = stringf("%s - %s", cell.name.c_str(), cell.title.c_str());
-		else title_line = cell.name;
-		string underline = "\n";
-		underline.insert(0, title_line.length(), '=');
-		fprintf(f, "%s\n", title_line.c_str());
-		fprintf(f, "%s\n", underline.c_str());
-
-		// help text, with cell def for links
-		fprintf(f, ".. cell:def:: %s\n", cell.name.c_str());
-		if (cell.title.length())
-			fprintf(f, "   :title: %s\n\n", cell.title.c_str());
-		else
-			fprintf(f, "   :title: %s\n\n", cell.name.c_str());
-		std::stringstream ss;
-		ss << cell.desc;
-		for (std::string line; std::getline(ss, line, '\n');) {
-			fprintf(f, "   %s\n", line.c_str());
-		}
-
-		// properties
-		fprintf(f, "\nProperties");
-		fprintf(f, "\n----------\n\n");
-		dict<string, bool> prop_dict = {
-			{"is_evaluable", ct.is_evaluable},
-			{"is_combinatorial", ct.is_combinatorial},
-			{"is_synthesizable", ct.is_synthesizable},
-		};
-		for (auto &it : prop_dict) {
-			fprintf(f, "- %s: %s\n", it.first.c_str(), it.second ? "true" : "false");
-		}
-
-		// source code
-		fprintf(f, "\nSimulation model (Verilog)");
-		fprintf(f, "\n--------------------------\n\n");
-		fprintf(f, ".. code-block:: verilog\n");
-		fprintf(f, "   :caption: %s\n\n", cell.source.c_str());
-		std::stringstream ss2;
-		ss2 << cell.code;
-		for (std::string line; std::getline(ss2, line, '\n');) {
-			fprintf(f, "   %s\n", line.c_str());
-		}
-
-		// footer
-		fprintf(f, "\n.. note::\n\n");
-		fprintf(f, "   This page was auto-generated from the output of\n");
-		fprintf(f, "   ``help %s``.\n", cell.name.c_str());
-
-		// close
-		fclose(f);
+		json.end_object();
+		return raise_error;
 	}
 	bool dump_cells_json(PrettyJson &json) {
 		// init json
 		json.begin_object();
 		json.entry("version", "Yosys internal cells");
-		json.entry("generator", yosys_version_str);
+		json.entry("generator", yosys_maybe_version());
 
 		dict<string, vector<string>> groups;
 		dict<string, pair<SimHelper, CellType>> cells;
@@ -1056,11 +1045,7 @@ struct HelpPass : public Pass {
 						log("=");
 					log("\n");
 					it.second->help();
-					if (it.second->experimental_flag) {
-						log("\n");
-						log("WARNING: THE '%s' COMMAND IS EXPERIMENTAL.\n", it.first.c_str());
-						log("\n");
-					}
+					log_warning_flags(it.second);
 				}
 			}
 			else if (args[1] == "-cells") {
@@ -1074,44 +1059,9 @@ struct HelpPass : public Pass {
 				log("\n");
 				return;
 			}
-			// this option is undocumented as it is for internal use only
-			else if (args[1] == "-write-rst-command-reference-manual") {
-				for (auto &it : pass_register) {
-					std::ostringstream buf;
-					log_streams.push_back(&buf);
-					it.second->help();
-					if (it.second->experimental_flag) {
-						log("\n");
-						log("WARNING: THE '%s' COMMAND IS EXPERIMENTAL.\n", it.first.c_str());
-						log("\n");
-					}
-					log_streams.pop_back();
-					write_cmd_rst(it.first, it.second->short_help, buf.str());
-				}
-			}
-			// this option is also undocumented as it is for internal use only
-			else if (args[1] == "-write-rst-cells-manual") {
-				bool raise_error = false;
-				for (auto &it : yosys_celltypes.cell_types) {
-					auto name = it.first.str();
-					if (cell_help_messages.contains(name)) {
-						write_cell_rst(cell_help_messages.get(name), it.second);
-					} else {
-						log("ERROR: Missing cell help for cell '%s'.\n", name.c_str());
-						raise_error |= true;
-					}
-				}
-				if (raise_error) {
-					log_error("One or more cells defined in celltypes.h are missing help documentation.\n");
-				}
-			}
 			else if (pass_register.count(args[1])) {
 				pass_register.at(args[1])->help();
-				if (pass_register.at(args[1])->experimental_flag) {
-					log("\n");
-					log("WARNING: THE '%s' COMMAND IS EXPERIMENTAL.\n", args[1].c_str());
-					log("\n");
-				}
+				log_warning_flags(pass_register.at(args[1]));
 			}
 			else if (cell_help_messages.contains(args[1])) {
 				auto help_cell = cell_help_messages.get(args[1]);
@@ -1140,7 +1090,17 @@ struct HelpPass : public Pass {
 				log("No such command or cell type: %s\n", args[1].c_str());
 			return;
 		} else if (args.size() == 3) {
-			if (args[1] == "-dump-cells-json") {
+			// this option is undocumented as it is for internal use only
+			if (args[1] == "-dump-cmds-json") {
+				PrettyJson json;
+				if (!json.write_to_file(args[2]))
+					log_error("Can't open file `%s' for writing: %s\n", args[2].c_str(), strerror(errno));
+				if (dump_cmds_json(json)) {
+					log_abort();
+				}
+			}
+			// this option is undocumented as it is for internal use only
+			else if (args[1] == "-dump-cells-json") {
 				PrettyJson json;
 				if (!json.write_to_file(args[2]))
 					log_error("Can't open file `%s' for writing: %s\n", args[2].c_str(), strerror(errno));
@@ -1148,6 +1108,8 @@ struct HelpPass : public Pass {
 					log_error("One or more cells defined in celltypes.h are missing help documentation.\n");
 				}
 			}
+			else
+				log("Unknown help command: `%s %s'\n", args[1].c_str(), args[2].c_str());
 			return;
 		}
 
