@@ -73,6 +73,23 @@ static char * pMcncGenlib[] = {
 };
 */
 
+// internal version of genlib library
+static char * pAndGenlib[] = {
+    "GATE zero    0   O=CONST0;\n",
+    "GATE one     0   O=CONST1;\n",
+    "GATE buf     1   O=a;              PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE inv     1   O=!a;             PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE and00   1   O=a*b;            PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE and01   1   O=a*!b;           PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE and10   1   O=!a*b;           PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE and11   1   O=!a*!b;          PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nand00  1   O=!(a*b);         PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nand01  1   O=!(a*!b);        PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nand10  1   O=!(!a*b);        PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nand11  1   O=!(!a*!b);       PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    NULL
+};
+
 // internal version of genlib library
 static char * pSimpleGenlib[] = {
     "GATE zero    0   O=CONST0;\n",
@@ -90,6 +107,27 @@ static char * pSimpleGenlib[] = {
     NULL
 };
 
+// internal version of genlib library
+static char * pSimpleGenlib2[] = {
+    "GATE zero    0   O=CONST0;\n",
+    "GATE one     0   O=CONST1;\n",
+    "GATE buf     4   O=a;              PIN * NONINV  1 999 1.0 0.0 1.0 0.0\n",
+    "GATE inv     2   O=!a;             PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nand2   4   O=!(a*b);         PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nand3   6   O=!(a*b*c);       PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nand4   8   O=!(a*b*c*d);     PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nor2    4   O=!(a+b);         PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nor3    6   O=!(a+b+c);       PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE nor4    6   O=!(a+b+c+d);     PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE aoi21   6   O=!(a*b+c);       PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE oai21   6   O=!((a+b)*c);     PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE aoi22   8   O=!(a*b+c*d);     PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE oai22   8   O=!((a+b)*(c+d)); PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE aoi211  8   O=!(a*b+c+d);     PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    "GATE oai211  8   O=!((a+b)*c*d);   PIN * INV     1 999 1.0 0.0 1.0 0.0\n",
+    NULL
+};
+
 ////////////////////////////////////////////////////////////////////////
 ///                     FUNCTION DEFINITIONS                         ///
 ////////////////////////////////////////////////////////////////////////
@@ -105,6 +143,20 @@ static char * pSimpleGenlib[] = {
   SeeAlso     []
 
 ***********************************************************************/
+void Mio_IntallAndLibrary()
+{
+    extern Mio_Library_t * Mio_LibraryReadBuffer( char * pBuffer, int fExtendedFormat, st__table * tExcludeGate, int nFaninLimit, int fVerbose );
+    Vec_Str_t * vLibStr = Vec_StrAlloc( 1000 );
+    for ( int i = 0; pAndGenlib[i]; i++ )
+        Vec_StrAppend( vLibStr, pAndGenlib[i] );
+    Vec_Str_t * vLibStr2 = Vec_StrDup( vLibStr );
+    Vec_StrAppend( vLibStr2, ".end\n" );
+    Vec_StrPush( vLibStr2, '\0' );
+    Vec_StrPush( vLibStr, '\0' );
+    Mio_UpdateGenlib2( vLibStr, vLibStr2, "and.genlib", 0 );
+    Vec_StrFree( vLibStr );    
+    Vec_StrFree( vLibStr2 );
+}
 void Mio_IntallSimpleLibrary()
 {
     extern Mio_Library_t * Mio_LibraryReadBuffer( char * pBuffer, int fExtendedFormat, st__table * tExcludeGate, int nFaninLimit, int fVerbose );
@@ -118,6 +170,19 @@ void Mio_IntallSimpleLibrary()
     Mio_UpdateGenlib( pLib );
     Vec_StrFree( vLibStr );
 }
+void Mio_IntallSimpleLibrary2()
+{
+    extern Mio_Library_t * Mio_LibraryReadBuffer( char * pBuffer, int fExtendedFormat, st__table * tExcludeGate, int nFaninLimit, int fVerbose );
+    Mio_Library_t * pLib;  int i;
+    Vec_Str_t * vLibStr = Vec_StrAlloc( 1000 );
+    for ( i = 0; pSimpleGenlib2[i]; i++ )
+        Vec_StrAppend( vLibStr, pSimpleGenlib2[i] );
+    Vec_StrPush( vLibStr, '\0' );
+    pLib = Mio_LibraryReadBuffer( Vec_StrArray(vLibStr), 0, NULL, 0, 0 );
+    Mio_LibrarySetName( pLib, Abc_UtilStrsav("simple2.genlib") );
+    Mio_UpdateGenlib( pLib );
+    Vec_StrFree( vLibStr );
+}
 
 /**Function*************************************************************
 
